#include "libdnf.hpp"
#include <array>
#include <cstddef>
#include <cstdint>
#include <exception>
#include <memory>
#include <new>
#include <string>
#include <type_traits>
#include <utility>

namespace rust {
inline namespace cxxbridge1 {
// #include "rust/cxx.h"

struct unsafe_bitcopy_t;

namespace {
template <typename T>
class impl;
} // namespace

#ifndef CXXBRIDGE1_RUST_STRING
#define CXXBRIDGE1_RUST_STRING
class String final {
public:
  String() noexcept;
  String(const String &) noexcept;
  String(String &&) noexcept;
  ~String() noexcept;

  String(const std::string &);
  String(const char *);
  String(const char *, std::size_t);
  String(const char16_t *);
  String(const char16_t *, std::size_t);
#if __cplusplus >= 202002L
  String(const char8_t *s);
  String(const char8_t *s, std::size_t len);
#endif

  static String lossy(const std::string &) noexcept;
  static String lossy(const char *) noexcept;
  static String lossy(const char *, std::size_t) noexcept;
  static String lossy(const char16_t *) noexcept;
  static String lossy(const char16_t *, std::size_t) noexcept;

  String &operator=(const String &) & noexcept;
  String &operator=(String &&) & noexcept;

  explicit operator std::string() const;

  const char *data() const noexcept;
  std::size_t size() const noexcept;
  std::size_t length() const noexcept;
  bool empty() const noexcept;

  const char *c_str() noexcept;

  std::size_t capacity() const noexcept;
  void reserve(size_t new_cap) noexcept;

  using iterator = char *;
  iterator begin() noexcept;
  iterator end() noexcept;

  using const_iterator = const char *;
  const_iterator begin() const noexcept;
  const_iterator end() const noexcept;
  const_iterator cbegin() const noexcept;
  const_iterator cend() const noexcept;

  bool operator==(const String &) const noexcept;
  bool operator!=(const String &) const noexcept;
  bool operator<(const String &) const noexcept;
  bool operator<=(const String &) const noexcept;
  bool operator>(const String &) const noexcept;
  bool operator>=(const String &) const noexcept;

  void swap(String &) noexcept;

  String(unsafe_bitcopy_t, const String &) noexcept;

private:
  struct lossy_t;
  String(lossy_t, const char *, std::size_t) noexcept;
  String(lossy_t, const char16_t *, std::size_t) noexcept;
  friend void swap(String &lhs, String &rhs) noexcept { lhs.swap(rhs); }

  std::array<std::uintptr_t, 3> repr;
};
#endif // CXXBRIDGE1_RUST_STRING

#ifndef CXXBRIDGE1_RUST_STR
#define CXXBRIDGE1_RUST_STR
class Str final {
public:
  Str() noexcept;
  Str(const String &) noexcept;
  Str(const std::string &);
  Str(const char *);
  Str(const char *, std::size_t);

  Str &operator=(const Str &) & noexcept = default;

  explicit operator std::string() const;

  const char *data() const noexcept;
  std::size_t size() const noexcept;
  std::size_t length() const noexcept;
  bool empty() const noexcept;

  Str(const Str &) noexcept = default;
  ~Str() noexcept = default;

  using iterator = const char *;
  using const_iterator = const char *;
  const_iterator begin() const noexcept;
  const_iterator end() const noexcept;
  const_iterator cbegin() const noexcept;
  const_iterator cend() const noexcept;

  bool operator==(const Str &) const noexcept;
  bool operator!=(const Str &) const noexcept;
  bool operator<(const Str &) const noexcept;
  bool operator<=(const Str &) const noexcept;
  bool operator>(const Str &) const noexcept;
  bool operator>=(const Str &) const noexcept;

  void swap(Str &) noexcept;

private:
  class uninit;
  Str(uninit) noexcept;
  friend impl<Str>;

  std::array<std::uintptr_t, 2> repr;
};
#endif // CXXBRIDGE1_RUST_STR

#ifndef CXXBRIDGE1_RUST_BITCOPY_T
#define CXXBRIDGE1_RUST_BITCOPY_T
struct unsafe_bitcopy_t final {
  explicit unsafe_bitcopy_t() = default;
};
#endif // CXXBRIDGE1_RUST_BITCOPY_T

#ifndef CXXBRIDGE1_RUST_BITCOPY
#define CXXBRIDGE1_RUST_BITCOPY
constexpr unsafe_bitcopy_t unsafe_bitcopy{};
#endif // CXXBRIDGE1_RUST_BITCOPY

#ifndef CXXBRIDGE1_IS_COMPLETE
#define CXXBRIDGE1_IS_COMPLETE
namespace detail {
namespace {
template <typename T, typename = std::size_t>
struct is_complete : std::false_type {};
template <typename T>
struct is_complete<T, decltype(sizeof(T))> : std::true_type {};
} // namespace
} // namespace detail
#endif // CXXBRIDGE1_IS_COMPLETE

namespace repr {
struct PtrLen final {
  void *ptr;
  ::std::size_t len;
};
} // namespace repr

namespace detail {
class Fail final {
  ::rust::repr::PtrLen &throw$;
public:
  Fail(::rust::repr::PtrLen &throw$) noexcept : throw$(throw$) {}
  void operator()(char const *) noexcept;
  void operator()(std::string const &) noexcept;
};
} // namespace detail

namespace {
template <bool> struct deleter_if {
  template <typename T> void operator()(T *) {}
};

template <> struct deleter_if<true> {
  template <typename T> void operator()(T *ptr) { ptr->~T(); }
};
} // namespace
} // namespace cxxbridge1

namespace behavior {
class missing {};
missing trycatch(...);

template <typename Try, typename Fail>
static typename ::std::enable_if<
    ::std::is_same<decltype(trycatch(::std::declval<Try>(), ::std::declval<Fail>())),
                 missing>::value>::type
trycatch(Try &&func, Fail &&fail) noexcept try {
  func();
} catch (::std::exception const &e) {
  fail(e.what());
}
} // namespace behavior
} // namespace rust

namespace dnfcxx {
  struct Nevra;
  using DnfPackage = ::dnfcxx::DnfPackage;
  using DnfRepo = ::dnfcxx::DnfRepo;
  using DnfSack = ::dnfcxx::DnfSack;
}

namespace dnfcxx {
#ifndef CXXBRIDGE1_STRUCT_dnfcxx$Nevra
#define CXXBRIDGE1_STRUCT_dnfcxx$Nevra
struct Nevra final {
  ::rust::String name;
  ::std::uint64_t epoch;
  ::rust::String version;
  ::rust::String release;
  ::rust::String arch;

  using IsRelocatable = ::std::true_type;
};
#endif // CXXBRIDGE1_STRUCT_dnfcxx$Nevra

extern "C" {
::dnfcxx::FFIDnfPackage *dnfcxx$cxxbridge1$DnfPackage$get_ref(::dnfcxx::DnfPackage &self) noexcept {
  ::dnfcxx::FFIDnfPackage &(::dnfcxx::DnfPackage::*get_ref$)() = &::dnfcxx::DnfPackage::get_ref;
  return &(self.*get_ref$)();
}

void dnfcxx$cxxbridge1$DnfPackage$get_nevra(::dnfcxx::DnfPackage &self, ::rust::String *return$) noexcept {
  ::rust::String (::dnfcxx::DnfPackage::*get_nevra$)() = &::dnfcxx::DnfPackage::get_nevra;
  new (return$) ::rust::String((self.*get_nevra$)());
}

void dnfcxx$cxxbridge1$DnfPackage$get_name(::dnfcxx::DnfPackage &self, ::rust::String *return$) noexcept {
  ::rust::String (::dnfcxx::DnfPackage::*get_name$)() = &::dnfcxx::DnfPackage::get_name;
  new (return$) ::rust::String((self.*get_name$)());
}

void dnfcxx$cxxbridge1$DnfPackage$get_evr(::dnfcxx::DnfPackage &self, ::rust::String *return$) noexcept {
  ::rust::String (::dnfcxx::DnfPackage::*get_evr$)() = &::dnfcxx::DnfPackage::get_evr;
  new (return$) ::rust::String((self.*get_evr$)());
}

void dnfcxx$cxxbridge1$DnfPackage$get_arch(::dnfcxx::DnfPackage &self, ::rust::String *return$) noexcept {
  ::rust::String (::dnfcxx::DnfPackage::*get_arch$)() = &::dnfcxx::DnfPackage::get_arch;
  new (return$) ::rust::String((self.*get_arch$)());
}

void dnfcxx$cxxbridge1$DnfPackage$get_sourcerpm(::dnfcxx::DnfPackage &self, ::rust::String *return$) noexcept {
  ::rust::String (::dnfcxx::DnfPackage::*get_sourcerpm$)() = &::dnfcxx::DnfPackage::get_sourcerpm;
  new (return$) ::rust::String((self.*get_sourcerpm$)());
}

::dnfcxx::DnfPackage *dnfcxx$cxxbridge1$dnf_package_from_ptr(::dnfcxx::FFIDnfPackage *pkg) noexcept {
  ::std::unique_ptr<::dnfcxx::DnfPackage> (*dnf_package_from_ptr$)(::dnfcxx::FFIDnfPackage *) = ::dnfcxx::dnf_package_from_ptr;
  return dnf_package_from_ptr$(pkg).release();
}

::dnfcxx::FFIDnfRepo *dnfcxx$cxxbridge1$DnfRepo$get_ref(::dnfcxx::DnfRepo &self) noexcept {
  ::dnfcxx::FFIDnfRepo &(::dnfcxx::DnfRepo::*get_ref$)() = &::dnfcxx::DnfRepo::get_ref;
  return &(self.*get_ref$)();
}

void dnfcxx$cxxbridge1$DnfRepo$get_id(::dnfcxx::DnfRepo &self, ::rust::String *return$) noexcept {
  ::rust::String (::dnfcxx::DnfRepo::*get_id$)() = &::dnfcxx::DnfRepo::get_id;
  new (return$) ::rust::String((self.*get_id$)());
}

::std::uint64_t dnfcxx$cxxbridge1$DnfRepo$get_timestamp_generated(::dnfcxx::DnfRepo &self) noexcept {
  ::std::uint64_t (::dnfcxx::DnfRepo::*get_timestamp_generated$)() = &::dnfcxx::DnfRepo::get_timestamp_generated;
  return (self.*get_timestamp_generated$)();
}

::dnfcxx::DnfRepo *dnfcxx$cxxbridge1$dnf_repo_from_ptr(::dnfcxx::FFIDnfRepo *pkg) noexcept {
  ::std::unique_ptr<::dnfcxx::DnfRepo> (*dnf_repo_from_ptr$)(::dnfcxx::FFIDnfRepo *) = ::dnfcxx::dnf_repo_from_ptr;
  return dnf_repo_from_ptr$(pkg).release();
}

::dnfcxx::FFIDnfSack *dnfcxx$cxxbridge1$DnfSack$get_ref(::dnfcxx::DnfSack &self) noexcept {
  ::dnfcxx::FFIDnfSack &(::dnfcxx::DnfSack::*get_ref$)() = &::dnfcxx::DnfSack::get_ref;
  return &(self.*get_ref$)();
}

::rust::repr::PtrLen dnfcxx$cxxbridge1$DnfSack$add_cmdline_package(::dnfcxx::DnfSack &self, ::rust::String const *filename, ::dnfcxx::DnfPackage **return$) noexcept {
  ::std::unique_ptr<::dnfcxx::DnfPackage> (::dnfcxx::DnfSack::*add_cmdline_package$)(::rust::String) = &::dnfcxx::DnfSack::add_cmdline_package;
  ::rust::repr::PtrLen throw$;
  ::rust::behavior::trycatch(
      [&] {
        new (return$) ::dnfcxx::DnfPackage *((self.*add_cmdline_package$)(::rust::String(::rust::unsafe_bitcopy, *filename)).release());
        throw$.ptr = nullptr;
      },
      ::rust::detail::Fail(throw$));
  return throw$;
}

::dnfcxx::DnfSack *dnfcxx$cxxbridge1$dnf_sack_new() noexcept {
  ::std::unique_ptr<::dnfcxx::DnfSack> (*dnf_sack_new$)() = ::dnfcxx::dnf_sack_new;
  return dnf_sack_new$().release();
}

::rust::repr::PtrLen dnfcxx$cxxbridge1$hy_split_nevra(::rust::Str nevra, ::dnfcxx::Nevra *return$) noexcept {
  ::dnfcxx::Nevra (*hy_split_nevra$)(::rust::Str) = ::dnfcxx::hy_split_nevra;
  ::rust::repr::PtrLen throw$;
  ::rust::behavior::trycatch(
      [&] {
        new (return$) ::dnfcxx::Nevra(hy_split_nevra$(nevra));
        throw$.ptr = nullptr;
      },
      ::rust::detail::Fail(throw$));
  return throw$;
}
} // extern "C"
} // namespace dnfcxx

extern "C" {
static_assert(::rust::detail::is_complete<::dnfcxx::DnfPackage>::value, "definition of DnfPackage is required");
static_assert(sizeof(::std::unique_ptr<::dnfcxx::DnfPackage>) == sizeof(void *), "");
static_assert(alignof(::std::unique_ptr<::dnfcxx::DnfPackage>) == alignof(void *), "");
void cxxbridge1$unique_ptr$dnfcxx$DnfPackage$null(::std::unique_ptr<::dnfcxx::DnfPackage> *ptr) noexcept {
  ::new (ptr) ::std::unique_ptr<::dnfcxx::DnfPackage>();
}
void cxxbridge1$unique_ptr$dnfcxx$DnfPackage$raw(::std::unique_ptr<::dnfcxx::DnfPackage> *ptr, ::dnfcxx::DnfPackage *raw) noexcept {
  ::new (ptr) ::std::unique_ptr<::dnfcxx::DnfPackage>(raw);
}
::dnfcxx::DnfPackage const *cxxbridge1$unique_ptr$dnfcxx$DnfPackage$get(::std::unique_ptr<::dnfcxx::DnfPackage> const &ptr) noexcept {
  return ptr.get();
}
::dnfcxx::DnfPackage *cxxbridge1$unique_ptr$dnfcxx$DnfPackage$release(::std::unique_ptr<::dnfcxx::DnfPackage> &ptr) noexcept {
  return ptr.release();
}
void cxxbridge1$unique_ptr$dnfcxx$DnfPackage$drop(::std::unique_ptr<::dnfcxx::DnfPackage> *ptr) noexcept {
  ::rust::deleter_if<::rust::detail::is_complete<::dnfcxx::DnfPackage>::value>{}(ptr);
}

static_assert(::rust::detail::is_complete<::dnfcxx::DnfRepo>::value, "definition of DnfRepo is required");
static_assert(sizeof(::std::unique_ptr<::dnfcxx::DnfRepo>) == sizeof(void *), "");
static_assert(alignof(::std::unique_ptr<::dnfcxx::DnfRepo>) == alignof(void *), "");
void cxxbridge1$unique_ptr$dnfcxx$DnfRepo$null(::std::unique_ptr<::dnfcxx::DnfRepo> *ptr) noexcept {
  ::new (ptr) ::std::unique_ptr<::dnfcxx::DnfRepo>();
}
void cxxbridge1$unique_ptr$dnfcxx$DnfRepo$raw(::std::unique_ptr<::dnfcxx::DnfRepo> *ptr, ::dnfcxx::DnfRepo *raw) noexcept {
  ::new (ptr) ::std::unique_ptr<::dnfcxx::DnfRepo>(raw);
}
::dnfcxx::DnfRepo const *cxxbridge1$unique_ptr$dnfcxx$DnfRepo$get(::std::unique_ptr<::dnfcxx::DnfRepo> const &ptr) noexcept {
  return ptr.get();
}
::dnfcxx::DnfRepo *cxxbridge1$unique_ptr$dnfcxx$DnfRepo$release(::std::unique_ptr<::dnfcxx::DnfRepo> &ptr) noexcept {
  return ptr.release();
}
void cxxbridge1$unique_ptr$dnfcxx$DnfRepo$drop(::std::unique_ptr<::dnfcxx::DnfRepo> *ptr) noexcept {
  ::rust::deleter_if<::rust::detail::is_complete<::dnfcxx::DnfRepo>::value>{}(ptr);
}

static_assert(::rust::detail::is_complete<::dnfcxx::DnfSack>::value, "definition of DnfSack is required");
static_assert(sizeof(::std::unique_ptr<::dnfcxx::DnfSack>) == sizeof(void *), "");
static_assert(alignof(::std::unique_ptr<::dnfcxx::DnfSack>) == alignof(void *), "");
void cxxbridge1$unique_ptr$dnfcxx$DnfSack$null(::std::unique_ptr<::dnfcxx::DnfSack> *ptr) noexcept {
  ::new (ptr) ::std::unique_ptr<::dnfcxx::DnfSack>();
}
void cxxbridge1$unique_ptr$dnfcxx$DnfSack$raw(::std::unique_ptr<::dnfcxx::DnfSack> *ptr, ::dnfcxx::DnfSack *raw) noexcept {
  ::new (ptr) ::std::unique_ptr<::dnfcxx::DnfSack>(raw);
}
::dnfcxx::DnfSack const *cxxbridge1$unique_ptr$dnfcxx$DnfSack$get(::std::unique_ptr<::dnfcxx::DnfSack> const &ptr) noexcept {
  return ptr.get();
}
::dnfcxx::DnfSack *cxxbridge1$unique_ptr$dnfcxx$DnfSack$release(::std::unique_ptr<::dnfcxx::DnfSack> &ptr) noexcept {
  return ptr.release();
}
void cxxbridge1$unique_ptr$dnfcxx$DnfSack$drop(::std::unique_ptr<::dnfcxx::DnfSack> *ptr) noexcept {
  ::rust::deleter_if<::rust::detail::is_complete<::dnfcxx::DnfSack>::value>{}(ptr);
}
} // extern "C"
